# Gridlock - a collection of grid-based strategy games
# Copyright (c) 2009 Brian Nenninger (brian@dozingcatsoftware.com)
# Gridlock comes with ABSOLUTELY NO WARRANTY, and is released under version 3 of the GNU GPL.

import time
import random

# from http://aima.cs.berkeley.edu/python/games.html
WIN_UTILITY = 2**30

# returns (move, next_state, utility) tuple
def alphabeta(eval_fn, state, depth, alpha, beta, nummoves, randomize=False):
    best = None
    # randomize moves at the top of the search tree, to avoid determinism
    moveiter = state.successors_for_ai()
    if randomize: 
        moveiter = list(moveiter)
        random.shuffle(moveiter)
    
    for move, nextstate in moveiter:
        utility = 0
        # stop recursion if game is over or we're at depth 1
        if nextstate.is_game_over():
            winner = nextstate.winning_player()
            # extra points for winning faster, minus points for losing faster
            if winner==state.current_player:
                utility = WIN_UTILITY+depth
            elif winner!=0:
                utility = -WIN_UTILITY-depth
            else:
                # tie; use normal evaluation?
                utility = -eval_fn(nextstate)
        elif depth<=1:
            utility = -eval_fn(nextstate)
        else:
            # continue down the tree, swapping alpha and beta and negating utility since players are switched
            nextmove, n2state, opputility = alphabeta(eval_fn, nextstate, depth-1, -beta, -alpha, nummoves)
            utility = -opputility # since we just computed for opponent
            
        if utility>beta: 
            utility = beta
        if best is None or utility>alpha:
            best = (move, nextstate, utility)
        if utility>alpha:
            alpha = utility
        if nummoves: nummoves[0]+=1
        if alpha >= beta:
            break
    #print "d="+str(depth)+", returning move:"+str(best[0])+", alpha="+str(alpha)
    return best
    
class GenericAI(object):
    def __init__(self, params):
        self.params = params
        self.depth = params.get('depth',2)
        
    def best_move(self, state):
        nmoves = [0]
        t1 = time.time()
        move, nextstate, utility = alphabeta(self.game_utility, state, self.depth, 
                                            -WIN_UTILITY-100, WIN_UTILITY+100, nmoves, True)
        tdiff = max(time.time()-t1, 1e-6)
        #print('AI move: %s, examined %d moves in %0.3f (%0.0f/sec), utility=%d' % (str(move), nmoves[0], tdiff, nmoves[0]/tdiff, utility))
        return (move, nextstate, nmoves[0], tdiff)
        
    def game_utility(self, game):
        # FIXME: this doesn't work well for 1-player games, see SetsAI
        pnum = game.current_player
        opnum = game.next_player()
        return game.player_score(pnum)-game.player_score(opnum)
        